/* 2007-08-24 */

/*
Copyright (c) 2005 JSON.org

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

The Software shall be used for Good, not Evil.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

/*
Callbacks, comments, Unicode handling by Jean Gressmann (jean@0x42.de), 2007-2009.

For the added features the license above applies also.

Changelog:
2009-09-29
Make code object oriented by Dolphin.

2009-05-17 
Incorporated benrudiak@googlemail.com fix for UTF16 decoding.

2009-05-14 
Fixed float parsing bug related to a locale being set that didn't
use '.' as decimal point character (charles@transmissionbt.com).

2008-10-14 
Renamed states.IN to states.IT to avoid name clash which IN macro
defined in windef.h (alexey.pelykh@gmail.com)

2008-07-19 
Removed some duplicate code & debugging variable (charles@transmissionbt.com)

2008-05-28 
Made value structure ansi C compliant. This bug was report by 
trisk@acm.jhu.edu

2008-05-20 
Fixed bug reported by charles@transmissionbt.com where the switching 
from static to dynamic parse buffer did not copy the static parse 
buffer's content.
*/




#include "stdafx.h"
#include "JsonParser.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define __   -1     /* the universal error code */

typedef unsigned short UTF16;
namespace Json{
	/*
	Characters are mapped into these character classes. This allows for
	a significant reduction in the size of the state transition table.
	*/



	enum EClass {
		C_SPACE,  /* space */
		C_WHITE,  /* other whitespace */
		C_LCURB,  /* {  */
		C_RCURB,  /* } */
		C_LSQRB,  /* [ */
		C_RSQRB,  /* ] */
		C_COLON,  /* : */
		C_COMMA,  /* , */
		C_QUOTE,  /* " */
		C_BACKS,  /* \ */
		C_SLASH,  /* / */
		C_PLUS,   /* + */
		C_MINUS,  /* - */
		C_POINT,  /* . */
		C_ZERO ,  /* 0 */
		C_DIGIT,  /* 123456789 */
		C_LOW_A,  /* a */
		C_LOW_B,  /* b */
		C_LOW_C,  /* c */
		C_LOW_D,  /* d */
		C_LOW_E,  /* e */
		C_LOW_F,  /* f */
		C_LOW_L,  /* l */
		C_LOW_N,  /* n */
		C_LOW_R,  /* r */
		C_LOW_S,  /* s */
		C_LOW_T,  /* t */
		C_LOW_U,  /* u */
		C_ABCDF,  /* ABCDF */
		C_E,      /* E */
		C_ETC,    /* everything else */
		C_STAR,   /* * */   
		NR_CLASSES
	};

	static int ascii_class[128] = {
		/*
		This array maps the 128 ASCII characters into character classes.
		The remaining Unicode characters should be mapped to C_ETC.
		Non-whitespace control characters are errors.
		*/
		__,      __,      __,      __,      __,      __,      __,      __,
		__,      C_WHITE, C_WHITE, __,      __,      C_WHITE, __,      __,
		__,      __,      __,      __,      __,      __,      __,      __,
		__,      __,      __,      __,      __,      __,      __,      __,

		C_SPACE, C_ETC,   C_QUOTE, C_ETC,   C_ETC,   C_ETC,   C_ETC,   C_ETC,
		C_ETC,   C_ETC,   C_STAR,   C_PLUS,  C_COMMA, C_MINUS, C_POINT, C_SLASH,
		C_ZERO,  C_DIGIT, C_DIGIT, C_DIGIT, C_DIGIT, C_DIGIT, C_DIGIT, C_DIGIT,
		C_DIGIT, C_DIGIT, C_COLON, C_ETC,   C_ETC,   C_ETC,   C_ETC,   C_ETC,

		C_ETC,   C_ABCDF, C_ABCDF, C_ABCDF, C_ABCDF, C_E,     C_ABCDF, C_ETC,
		C_ETC,   C_ETC,   C_ETC,   C_ETC,   C_ETC,   C_ETC,   C_ETC,   C_ETC,
		C_ETC,   C_ETC,   C_ETC,   C_ETC,   C_ETC,   C_ETC,   C_ETC,   C_ETC,
		C_ETC,   C_ETC,   C_ETC,   C_LSQRB, C_BACKS, C_RSQRB, C_ETC,   C_ETC,

		C_ETC,   C_LOW_A, C_LOW_B, C_LOW_C, C_LOW_D, C_LOW_E, C_LOW_F, C_ETC,
		C_ETC,   C_ETC,   C_ETC,   C_ETC,   C_LOW_L, C_ETC,   C_LOW_N, C_ETC,
		C_ETC,   C_ETC,   C_LOW_R, C_LOW_S, C_LOW_T, C_LOW_U, C_ETC,   C_ETC,
		C_ETC,   C_ETC,   C_ETC,   C_LCURB, C_ETC,   C_RCURB, C_ETC,   C_ETC
	};


	/*
	The state codes.
	*/
	enum EState {
		GO,  /* start    */
		OK,  /* ok       */
		OB,  /* object   */
		KE,  /* key      */
		CO,  /* colon    */
		VA,  /* value    */
		AR,  /* array    */
		ST,  /* string   */
		ES,  /* escape   */
		U1,  /* u1       */
		U2,  /* u2       */
		U3,  /* u3       */
		U4,  /* u4       */
		MI,  /* minus    */
		ZE,  /* zero     */
		IT,  /* integer  */
		FR,  /* fraction */
		E1,  /* e        */
		E2,  /* ex       */
		E3,  /* exp      */
		T1,  /* tr       */
		T2,  /* tru      */
		T3,  /* true     */
		F1,  /* fa       */
		F2,  /* fal      */
		F3,  /* fals     */
		F4,  /* false    */
		N1,  /* nu       */
		N2,  /* nul      */
		N3,  /* null     */
		C1,  /* /        */
		C2,  /* / *     */
		C3,  /* *        */
		FX,  /* *.* *eE* */
		D1,  /* second UTF-16 character decoding started by \ */
		D2,  /* second UTF-16 character proceeded by u */
		NR_STATES
	};

	enum EAction
	{
		CB = -10, /* comment begin */
		CE = -11, /* comment end */
		FA = -12, /* false */
		TR = -13, /* false */
		NU = -14, /* null */
		DE = -15, /* double detected by exponent e E */
		DF = -16, /* double detected by fraction . */
		SB = -17, /* string begin */
		MX = -18, /* integer detected by minus */
		ZX = -19, /* integer detected by zero */
		IX = -20, /* integer detected by 1-9 */
		EX = -21, /* next char is escaped */
		UC = -22  /* Unicode character read */
	};


	static int state_transition_table[NR_STATES][NR_CLASSES] = {
		/*
		The state transition table takes the current state and the current symbol,
		and returns either a new state or an action. An action is represented as a
		negative number. A JSON text is accepted if at the end of the text the
		state is OK and if the mode is MODE_DONE.

		white                                      1-9                                   ABCDF  etc
		space |  {  }  [  ]  :  ,  "  \  /  +  -  .  0  |  a  b  c  d  e  f  l  n  r  s  t  u  |  E  |  * */
		/*start  GO*/ {GO,GO,-6,__,-5,__,__,__,__,__,CB,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__},
		/*ok     OK*/ {OK,OK,__,-8,__,-7,__,-3,__,__,CB,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__},
		/*object OB*/ {OB,OB,__,-9,__,__,__,__,SB,__,CB,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__},
		/*key    KE*/ {KE,KE,__,__,__,__,__,__,SB,__,CB,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__},
		/*colon  CO*/ {CO,CO,__,__,__,__,-2,__,__,__,CB,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__},
		/*value  VA*/ {VA,VA,-6,__,-5,__,__,__,SB,__,CB,__,MX,__,ZX,IX,__,__,__,__,__,FA,__,NU,__,__,TR,__,__,__,__,__},
		/*array  AR*/ {AR,AR,-6,__,-5,-7,__,__,SB,__,CB,__,MX,__,ZX,IX,__,__,__,__,__,FA,__,NU,__,__,TR,__,__,__,__,__},
		/*string ST*/ {ST,__,ST,ST,ST,ST,ST,ST,-4,EX,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST,ST},
		/*escape ES*/ {__,__,__,__,__,__,__,__,ST,ST,ST,__,__,__,__,__,__,ST,__,__,__,ST,__,ST,ST,__,ST,U1,__,__,__,__},
		/*u1     U1*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,U2,U2,U2,U2,U2,U2,U2,U2,__,__,__,__,__,__,U2,U2,__,__},
		/*u2     U2*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,U3,U3,U3,U3,U3,U3,U3,U3,__,__,__,__,__,__,U3,U3,__,__},
		/*u3     U3*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,U4,U4,U4,U4,U4,U4,U4,U4,__,__,__,__,__,__,U4,U4,__,__},
		/*u4     U4*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,UC,UC,UC,UC,UC,UC,UC,UC,__,__,__,__,__,__,UC,UC,__,__},
		/*minus  MI*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,ZE,IT,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__},
		/*zero   ZE*/ {OK,OK,__,-8,__,-7,__,-3,__,__,CB,__,__,DF,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__},
		/*int    IT*/ {OK,OK,__,-8,__,-7,__,-3,__,__,CB,__,__,DF,IT,IT,__,__,__,__,DE,__,__,__,__,__,__,__,__,DE,__,__},
		/*frac   FR*/ {OK,OK,__,-8,__,-7,__,-3,__,__,CB,__,__,__,FR,FR,__,__,__,__,E1,__,__,__,__,__,__,__,__,E1,__,__},
		/*e      E1*/ {__,__,__,__,__,__,__,__,__,__,__,E2,E2,__,E3,E3,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__},
		/*ex     E2*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,E3,E3,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__},
		/*exp    E3*/ {OK,OK,__,-8,__,-7,__,-3,__,__,__,__,__,__,E3,E3,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__},
		/*tr     T1*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,T2,__,__,__,__,__,__,__},
		/*tru    T2*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,T3,__,__,__,__},
		/*true   T3*/ {__,__,__,__,__,__,__,__,__,__,CB,__,__,__,__,__,__,__,__,__,OK,__,__,__,__,__,__,__,__,__,__,__},
		/*fa     F1*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,F2,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__},
		/*fal    F2*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,F3,__,__,__,__,__,__,__,__,__},
		/*fals   F3*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,F4,__,__,__,__,__,__},
		/*false  F4*/ {__,__,__,__,__,__,__,__,__,__,CB,__,__,__,__,__,__,__,__,__,OK,__,__,__,__,__,__,__,__,__,__,__},
		/*nu     N1*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,N2,__,__,__,__},
		/*nul    N2*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,N3,__,__,__,__,__,__,__,__,__},
		/*null   N3*/ {__,__,__,__,__,__,__,__,__,__,CB,__,__,__,__,__,__,__,__,__,__,__,OK,__,__,__,__,__,__,__,__,__},
		/*/      C1*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,C2},
		/*/*     C2*/ {C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C3},
		/**      C3*/ {C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,CE,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C2,C3},
		/*_.     FX*/ {OK,OK,__,-8,__,-7,__,-3,__,__,__,__,__,__,FR,FR,__,__,__,__,E1,__,__,__,__,__,__,__,__,E1,__,__},
		/*\      D1*/ {__,__,__,__,__,__,__,__,__,D2,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__},
		/*\      D2*/ {__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,__,U1,__,__,__,__},
	};


	/*
	These modes can be pushed on the stack.
	*/
	enum EMode {
		MODE_ARRAY = 1, 
		MODE_DONE = 2,  
		MODE_KEY = 3,   
		MODE_OBJECT = 4
	};

	bool CJsonParser::push(EMode mode)
	{
        if (top++ == stacktop)
            return false;
		*top = mode;
	    return true;
	}


	bool CJsonParser::pop(EMode mode)
	{
	    if (top < stack || *top != mode)
	        return false;
	    --top;
	    return true;
	}

#define parse_buffer_clear() \
	do {\
	parse_buffer_count = 0;\
	parse_buffer[0] = 0;\
	} while (0)

#define parse_buffer_pop_back_char()\
	do {\
	ASSERT(parse_buffer_count >= 1);\
	--parse_buffer_count;\
	parse_buffer[parse_buffer_count] = 0;\
	} while (0)    

	CJsonParser::~CJsonParser()
	{
		delete [] stack;
		delete [] parse_buffer;
	}

	CJsonParser::CJsonParser(sint_ptr _depth, sint_ptr init_buffer_capacity, bool _allow_comments)
		: depth(_depth)
		, parse_buffer_capacity(init_buffer_capacity)
		, allow_comments(_allow_comments)
		, type(T_NONE)
		, state(GO)
		, before_comment_state(0)
		, escaped(false)
		, comment(false)
		, utf16_high_surrogate(0)
	{
		/* We need to be able to push at least one object */
		if (depth <= 0) depth = 1;
		memset(callbacks, 0, sizeof(callbacks));

	    stack = new char[depth];
		top = stack - 1;
		stacktop = stack + depth;
	    /* set parser to start */
	    push(MODE_DONE);
		/* set up the parse buffer */
		parse_buffer = new char[parse_buffer_capacity];
		parse_buffer_clear();
	}
	void SetCallback(void* _callback_ctx){

	}
	void CJsonParser::grow_parse_buffer()
	{
		size_t bytes_to_allocate;
		parse_buffer_capacity *= 2;
		bytes_to_allocate = parse_buffer_capacity * sizeof(parse_buffer[0]);
		char* new_parse_buffer = new char[bytes_to_allocate];
		memcpy(new_parse_buffer, parse_buffer, parse_buffer_count);
		delete [] parse_buffer;
		parse_buffer = new_parse_buffer;
	}

#define parse_buffer_push_back_char(c)\
	do {\
	if (parse_buffer_count + 1 >= parse_buffer_capacity) grow_parse_buffer();\
	parse_buffer[parse_buffer_count++] = c;\
	parse_buffer[parse_buffer_count]   = 0;\
	} while (0)

#define assert_is_non_container_type() \
	ASSERT( \
	type == T_NULL || \
	type == T_FALSE || \
	type == T_TRUE || \
	type == T_FLOAT || \
	type == T_INTEGER || \
	type == T_STRING)


	bool CJsonParser::parse_parse_buffer()
	{
		Value value, *arg = NULL;

		if (type != T_NONE) {
			assert_is_non_container_type();

			switch(type) {
				case T_FLOAT:
				case T_INTEGER:
				case T_STRING:
					value.value = parse_buffer;
					value.length = parse_buffer_count;
				case T_NULL:
				case T_TRUE:
				case T_FALSE:
					arg = &value;
					value.type = type;
					break;
			}

			switch(*top){
				case MODE_ARRAY:
					if (callbacks[T_ITEM_END] && !(*callbacks[T_ITEM_END])(callback_ctx, arg))
						return false;
					break;
				case MODE_OBJECT:
					if (callbacks[T_VALUE] && !(*callbacks[T_VALUE])(callback_ctx, arg))
						return false;
					break;
			}
		}

		parse_buffer_clear();

		return true;
	}

#ifndef _WINNLS_
//#define IS_HIGH_SURROGATE(uc) (((uc) & 0xFC00) == 0xD800)
//#define IS_LOW_SURROGATE(uc)  (((uc) & 0xFC00) == 0xDC00)
//#define IS_HIGH_SURROGATE(wch) (((wch) >= HIGH_SURROGATE_START) && ((wch) <= HIGH_SURROGATE_END))
//#define IS_LOW_SURROGATE(wch)  (((wch) >= LOW_SURROGATE_START) && ((wch) <= LOW_SURROGATE_END))
//#define IS_SURROGATE_PAIR(hs, ls) (IS_HIGH_SURROGATE(hs) && IS_LOW_SURROGATE(ls)) 
#endif
#define DECODE_SURROGATE_PAIR(hi,lo) ((((hi) & 0x3FF) << 10) + ((lo) & 0x3FF) + 0x10000)
	static unsigned char utf8_lead_bits[4] = { 0x00, 0xC0, 0xE0, 0xF0 };

	bool CJsonParser::decode_unicode_char()
	{
		sint_ptr i;
		uint_ptr uc = 0;
		char* p;
		sint_ptr trail_bytes;

		ASSERT(parse_buffer_count >= 6);

		p = &parse_buffer[parse_buffer_count - 4];

		for (i = 12; i >= 0; i -= 4, ++p) {
			uint_ptr x = *p;

			if (x >= 'a') {
				x -= ('a' - 10);
			} else if (x >= 'A') {
				x -= ('A' - 10);
			} else {
				x &= ~0x30u;
			}

			ASSERT(x < 16);

			uc |= x << i;
		}

		/* clear UTF-16 char from buffer */
		parse_buffer_count -= 6;
		parse_buffer[parse_buffer_count] = 0;

		/* attempt decoding ... */
		if (utf16_high_surrogate) {
			if (IS_LOW_SURROGATE(uc)) {
				uc = DECODE_SURROGATE_PAIR(utf16_high_surrogate, uc);
				trail_bytes = 3;
				utf16_high_surrogate = 0;
			} else {
				/* high surrogate without a following low surrogate */
				return false;
			}
		} else {
			if (uc < 0x80) {
				trail_bytes = 0;
			} else if (uc < 0x800) {
				trail_bytes = 1;
			} else if (IS_HIGH_SURROGATE(uc)) {
				/* save the high surrogate and wait for the low surrogate */
				utf16_high_surrogate = (Json::UTF16)uc;
				return true;
			} else if (IS_LOW_SURROGATE(uc)) {
				/* low surrogate without a preceding high surrogate */
				return false;
			} else {
				trail_bytes = 2;
			}
		}

		parse_buffer[parse_buffer_count++] = (char) ((uc >> (trail_bytes * 6)) | utf8_lead_bits[trail_bytes]);

		for (i = trail_bytes * 6 - 6; i >= 0; i -= 6) {
			parse_buffer[parse_buffer_count++] = (char) (((uc >> i) & 0x3F) | 0x80);
		}

		parse_buffer[parse_buffer_count] = 0;

		return true;
	}

	bool CJsonParser::add_escaped_char_to_parse_buffer(int next_char)
	{
		escaped = false;
		/* remove the backslash */
		parse_buffer_pop_back_char();
		switch(next_char) {
			case 'b':
				parse_buffer_push_back_char('\b');
				break;
			case 'f':
				parse_buffer_push_back_char('\f');
				break;
			case 'n':
				parse_buffer_push_back_char('\n');
				break;
			case 'r':
				parse_buffer_push_back_char('\r');
				break;
			case 't':
				parse_buffer_push_back_char('\t');
				break;
			case '"':
				parse_buffer_push_back_char('"');
				break;
			case '\\':
				parse_buffer_push_back_char('\\');
				break;
			case '/':
				parse_buffer_push_back_char('/');
				break;
			case 'u':
				parse_buffer_push_back_char('\\');
				parse_buffer_push_back_char('u');
				break;
			default:
				return false;
		}

		return true;
	}

#define add_char_to_parse_buffer(next_char, next_class) \
	do { \
		if (escaped) { \
			if (!add_escaped_char_to_parse_buffer(next_char)) \
				return false; \
		} else if (!comment) { \
			if ((type != T_NONE) | !((next_class == C_SPACE) | (next_class == C_WHITE)) /* non-white-space */) { \
				parse_buffer_push_back_char((char)next_char); \
			} \
		} \
	} while (0)


#define assert_type_isnt_string_null_or_bool() \
	ASSERT(type != T_FALSE); \
	ASSERT(type != T_TRUE); \
	ASSERT(type != T_NULL); \
	ASSERT(type != T_STRING)


	bool CJsonParser::Next(unsigned int next_char)
	{
		/*
		After calling new_parser, call this function for each character (or
		partial character) in your JSON text. It can accept UTF-8, UTF-16, or
		UTF-32. It returns true if things are looking ok so far. If it rejects the
		text, it returns false.
		*/
		sint_ptr next_class, next_state;

		/*
		Determine the character's class.
		*/
		/*if (next_char < 0) {
			return false;
		}*/
		if (next_char >= 128) {
			next_class = C_ETC;
		} else {
			next_class = ascii_class[next_char];
			if (next_class <= __) {
				return false;
			}
		}

		add_char_to_parse_buffer(next_char, next_class);

		/*
		Get the next state from the state transition table.
		*/
		next_state = state_transition_table[state][next_class];
		if (next_state >= 0) {
			/*
			Change the state.
			*/
			state = next_state;
		} else {
			/*
			Or perform one of the actions.
			*/
			switch (next_state) {
				/* Unicode character */        
				case UC:
					if(!decode_unicode_char()) {
						return false;
					}
					/* check if we need to read a second UTF-16 char */
					if (utf16_high_surrogate) {
						state = D1;
					} else {
						state = ST;
					}
					break;
					/* escaped char */
				case EX:
					escaped = true;
					state = ES;
					break;
					/* integer detected by minus */
				case MX:
					type = T_INTEGER;
					state = MI;
					break;  
					/* integer detected by zero */            
				case ZX:
					type = T_INTEGER;
					state = ZE;
					break;  
					/* integer detected by 1-9 */            
				case IX:
					type = T_INTEGER;
					state = IT;
					break;  

					/* floating point number detected by exponent*/
				case DE:
					assert_type_isnt_string_null_or_bool();
					type = T_FLOAT;
					state = E1;
					break;   

					/* floating point number detected by fraction */
				case DF:
					assert_type_isnt_string_null_or_bool();
     				type = T_FLOAT;
					state = FX;
					break;   
					/* string begin " */
				case SB:
					parse_buffer_clear();
					ASSERT(type == T_NONE);
					type = T_STRING;
					state = ST;
					break;        

					/* n */
				case NU:
					ASSERT(type == T_NONE);
					type = T_NULL;
					state = N1;
					break;        
					/* f */
				case FA:
					ASSERT(type == T_NONE);
					type = T_FALSE;
					state = F1;
					break;        
					/* t */
				case TR:
					ASSERT(type == T_NONE);
					type = T_TRUE;
					state = T1;
					break;        

					/* closing comment */
				case CE:
					comment = false;
					ASSERT(parse_buffer_count == 0);
					ASSERT(type == T_NONE);
					state = before_comment_state;
					break;        

					/* opening comment  */
				case CB:
					if (!allow_comments) {
						return false;
					}
					parse_buffer_pop_back_char();
					if (!parse_parse_buffer()) {
						return false;
					}
					ASSERT(parse_buffer_count == 0);
					ASSERT(type != T_STRING);
					switch (*top) {
						case MODE_ARRAY:
						case MODE_OBJECT:   
							switch(state) {
								case VA:
								case AR:
									before_comment_state = state;
									break;
								default:
									before_comment_state = OK;
									break;
							}
							break;
						default:
							before_comment_state = state;
							break;
					}
					type = T_NONE;
					state = C1;
					comment = true;
					break;
					/* empty } */
				case -9:        
					parse_buffer_clear();
					if (callbacks[T_OBJECT_END] && !(*callbacks[T_OBJECT_END])(callback_ctx, NULL)) {
						return false;
					}
		            if (!pop(MODE_KEY)) {
		                return false;
		            }
					state = OK;
					break;

				/* } */ case -8:
					parse_buffer_pop_back_char();
					if (!parse_parse_buffer()) {
						return false;
					}
					if (callbacks[T_OBJECT_END] && !(*callbacks[T_OBJECT_END])(callback_ctx, NULL)) {
						return false;
					}
		            if (!pop(MODE_OBJECT))
		                return false;
					type = T_NONE;
					state = OK;
					break;

				/* ] */ case -7:
					parse_buffer_pop_back_char();
					if (!parse_parse_buffer()) {
						return false;
					}
					if (callbacks[T_ARRAY_END] && !(*callbacks[T_ARRAY_END])(callback_ctx, NULL)) {
						return false;
					}
		            if (!pop(MODE_ARRAY))
		                return false;
					type = T_NONE;
					state = OK;
					break;

				/* { */ case -6:
					parse_buffer_pop_back_char();
					if (callbacks[T_OBJECT_BEGIN] && !(*callbacks[T_OBJECT_BEGIN])(callback_ctx, NULL)) {
						return false;
					}
		            if (!push(MODE_KEY))
						return false;
					ASSERT(type == T_NONE);
					state = OB;
					break;

				/* [ */ case -5:
					parse_buffer_pop_back_char();
					if (callbacks[T_ARRAY_BEGIN] && !(*callbacks[T_ARRAY_BEGIN])(callback_ctx, NULL)) {
						return false;
					}
		            if (!push(MODE_ARRAY))
						return false;
					ASSERT(type == T_NONE);
					state = AR;
					break;

				/* string end " */ case -4:
					parse_buffer_pop_back_char();
					switch (*top) {
						case MODE_KEY:
							ASSERT(type == T_STRING);
							type = T_NONE;
							state = CO;
							if (callbacks[T_KEY]) {
								Value value;
								//value.type = type;
								value.value = parse_buffer;
								value.length = parse_buffer_count;
								if (!(*callbacks[T_KEY])(callback_ctx, &value)) {
									return false;
								}
							}
							parse_buffer_clear();
							break;
						case MODE_ARRAY:
						case MODE_OBJECT:
							ASSERT(type == T_STRING);
							if (!parse_parse_buffer()) {
								return false;
							}
							type = T_NONE;
							state = OK;
							break;
						default:
							return false;
					}
					break;

				/* , */ case -3:
					parse_buffer_pop_back_char();
					if (!parse_parse_buffer()) {
						return false;
					}
					switch (*top) {
						case MODE_OBJECT:
							/*
							A comma causes a flip from object mode to key mode.
							*/
			                if (!pop(MODE_OBJECT) || !push(MODE_KEY))
			                    return false;

							ASSERT(type != T_STRING);
							type = T_NONE;
							state = KE;
							break;
						case MODE_ARRAY:
							ASSERT(type != T_STRING);
							type = T_NONE;
							state = VA;
							break;
						default:
							return false;
					}
					break;

				/* : */ case -2:
					/*
					A colon causes a flip from key mode to object mode.
					*/
					parse_buffer_pop_back_char();
		            if (!pop(MODE_KEY) || !push(MODE_OBJECT)) {
		                return false;
		            }
					ASSERT(type == T_NONE);
					state = VA;
					break;
					/*
					Bad action.
					*/
				default:
					return false;
			}
		}
		return true;
	}


	bool CJsonParser::isDone(){
		return state == OK && pop(MODE_DONE);
	}

	/*
	int parser_is_legal_white_space_string(const char* s)
	{
	int c, char_class;

	if (s == NULL) {
	return false;
	}

	for (; *s; ++s) {   
	c = *s;

	if (c < 0 || c >= 128) {
	return false;
	}

	char_class = ascii_class[c];

	if (char_class != C_SPACE && char_class != C_WHITE) {
	return false;
	}
	}

	return true;
	}
	*/

}